--------------------------------------------------------------------------------
-- This file is part of the ParaNut project.
-- 
-- Copyright (C) 2022 Christian Meyer <christian.meyer@hs-augsburg.de>
--     Hochschule Augsburg, University of Applied Sciences
--
-- Redistribution and use in source and binary forms, with or without modification,
-- are permitted provided that the following conditions are met:
--
-- 1. Redistributions of source code must retain the above copyright notice, this 
--    list of conditions and the following disclaimer.
--
-- 2. Redistributions in binary form must reproduce the above copyright notice,
--    this list of conditions and the following disclaimer in the documentation and/or
--    other materials provided with the distribution.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
-- ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
-- WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
-- DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
-- ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
-- (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
-- LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
-- ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
-- (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
-- SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
--
-- Description:
--  Page Table Walker (PTW) module
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library paranut;
use paranut.memu_lib.all;

entity mptw_wrapper is
    port (
             clk            : in std_logic;
             reset          : in std_logic;
             -- directly connected to MemU
             memu_root_ppn  : in std_logic_vector (19 downto 0);
             -- WB ports...
             ptwwbi         : in busif_wishbone_in_type;
             ptwwbo         : out busif_wishbone_out_type;
             -- to BusIf...
             ptwi           : in ptw_in_type;
             ptwo           : out ptw_out_type;
             -- to TLB...
             tlbo            : out tlb_in_type;
             tlbi            : in tlb_out_type
         );
end mptw_wrapper;

architecture rtl of mptw_wrapper is

  component MPtw is
    port (
    clk : IN STD_LOGIC;
    reset : IN STD_LOGIC;
    wb_cyc_o : OUT STD_LOGIC;
    wb_stb_o : OUT STD_LOGIC;
    wb_we_o : OUT STD_LOGIC;
    wb_cti_o : OUT STD_LOGIC_VECTOR (2 downto 0);
    wb_bte_o : OUT STD_LOGIC_VECTOR (1 downto 0);
    wb_sel_o : OUT STD_LOGIC_VECTOR (3 downto 0);
    wb_adr_o : OUT STD_LOGIC_VECTOR (31 downto 0);
    wb_dat_o : OUT STD_LOGIC_VECTOR (31 downto 0);
    wb_ack_i : IN STD_LOGIC;
    wb_dat_i : IN STD_LOGIC_VECTOR (31 downto 0);
    memu_root_ppn : IN STD_LOGIC_VECTOR (19 downto 0);
    busif_req : IN STD_LOGIC;
    busif_ack : OUT STD_LOGIC;
    busif_virt_adr : IN STD_LOGIC_VECTOR (31 downto 0);
    busif_phys_adr : OUT STD_LOGIC_VECTOR (31 downto 0);
    busif_ac_r : OUT STD_LOGIC;
    busif_ac_w : OUT STD_LOGIC;
    busif_ac_x : OUT STD_LOGIC;
    busif_ac_u : OUT STD_LOGIC;
    busif_ac_d : OUT STD_LOGIC;
    busif_ac_a : OUT STD_LOGIC;
    tlb_req : OUT STD_LOGIC;
    tlb_wr : OUT STD_LOGIC;
    tlb_va_o : OUT STD_LOGIC_VECTOR (19 downto 0);
    tlb_pa_o : OUT STD_LOGIC_VECTOR (19 downto 0);
    tlb_superpage_o : OUT STD_LOGIC;
    tlb_ac_r_o : OUT STD_LOGIC;
    tlb_ac_w_o : OUT STD_LOGIC;
    tlb_ac_x_o : OUT STD_LOGIC;
    tlb_ac_u_o : OUT STD_LOGIC;
    tlb_ac_d_o : OUT STD_LOGIC;
    tlb_ac_a_o : OUT STD_LOGIC;
    tlb_superpage_i : IN STD_LOGIC;
    tlb_adr_i : IN STD_LOGIC_VECTOR (19 downto 0);
    tlb_hit : IN STD_LOGIC;
    tlb_miss : IN STD_LOGIC;
    tlb_ac_r_i : IN STD_LOGIC;
    tlb_ac_w_i : IN STD_LOGIC;
    tlb_ac_x_i : IN STD_LOGIC;
    tlb_ac_u_i : IN STD_LOGIC;
    tlb_ac_d_i : IN STD_LOGIC;
    tlb_ac_a_i : IN STD_LOGIC
    );
  end component;

begin

  HLSMptw : component MPtw
  port map (
      clk => clk,
      reset => reset,
      
      -- directly connected to MemU
      memu_root_ppn => memu_root_ppn,
      
      -- WB ports...
      wb_cyc_o => ptwwbo.cyc_o,
      wb_stb_o => ptwwbo.stb_o,
      wb_we_o => ptwwbo.we_o,
      wb_sel_o => ptwwbo.sel_o,
      wb_adr_o => ptwwbo.adr_o,
      wb_dat_o => ptwwbo.dat_o,
      wb_cti_o => ptwwbo.cti_o,
      wb_bte_o => ptwwbo.bte_o,

      wb_ack_i => ptwwbi.ack_i,
      wb_dat_i => ptwwbi.dat_i,

      -- to BusIf...
      busif_phys_adr => ptwo.busif_phys_adr,
      busif_ack => ptwo.busif_ack,
      busif_ac_r => ptwo.busif_ac_r,
      busif_ac_w => ptwo.busif_ac_w,
      busif_ac_x => ptwo.busif_ac_x,
      busif_ac_u => ptwo.busif_ac_u,
      busif_ac_d => ptwo.busif_ac_d,
      busif_ac_a => ptwo.busif_ac_a,

      -- from BusIf...
      busif_virt_adr => ptwi.busif_virt_adr,
      busif_req => ptwi.busif_req,
      
      -- to TLB...
      tlb_req => tlbo.ptw_req,
      tlb_wr => tlbo.ptw_wr,
      tlb_va_o => tlbo.ptw_va_i,
      tlb_pa_o => tlbo.ptw_pa_i,
      tlb_superpage_o => tlbo.ptw_superpage_i,
      tlb_ac_r_o => tlbo.ptw_ac_r_i,
      tlb_ac_w_o => tlbo.ptw_ac_w_i,
      tlb_ac_x_o => tlbo.ptw_ac_x_i,
      tlb_ac_u_o => tlbo.ptw_ac_u_i,
      tlb_ac_d_o => tlbo.ptw_ac_d_i,
      tlb_ac_a_o => tlbo.ptw_ac_a_i,

      -- from TLB...
      tlb_superpage_i =>  tlbi.ptw_superpage_o,
      tlb_adr_i => tlbi.ptw_adr_o,
      tlb_hit => tlbi.ptw_hit_o,
      tlb_miss => tlbi.ptw_miss_o,
      tlb_ac_r_i => tlbi.ptw_ac_r_o,
      tlb_ac_w_i => tlbi.ptw_ac_w_o,
      tlb_ac_x_i => tlbi.ptw_ac_x_o,
      tlb_ac_u_i => tlbi.ptw_ac_u_o,
      tlb_ac_d_i => tlbi.ptw_ac_a_o,
      tlb_ac_a_i => tlbi.ptw_ac_d_o
  );

end rtl;
